# Example of Importing CSV Files

The following example shows you how to import CSV data to **Nebula Graph** with **Nebula Importer**. In this example, **Nebula Graph** is installed with `Docker` and `Docker Compose`. We will walk you through the example by the following steps:

1. [Start your Nebula Graph services](#starting-nebula-graph-services)
2. [Create the schema for vertices and edges](#creating-the-schema-for-vertices-and-edges)
3. [Prepare the configuration file](#preparing-your-configuration-file)
4. [Prepare the CSV data](#preparing-the-csv-data)
5. [Import the CSV data](#importing-the-csv-data)

## Starting Nebula Graph Services

You can start your **Nebula Graph** services by the following steps:

1. On a command line interface, go to the `nebula-docker-compose` directory.
2. Execute the following command to start **Nebula Graph** services:

```bash
$ sudo docker-compose up -d
```

3. Execute the following command to pull the **Nebula Graph** image:

```bash
$ sudo docker pull vesoft/nebula-console:nightly
```

4. Execute the following command to connect to your **Nebula Graph** server:

```bash
$ sudo docker run --rm -ti --network=host vesoft/nebula-console:nightly --addr=127.0.0.1 --port=3699
```

**Note**: You must ensure your IP address and port number are configured correctly.

## Creating the Schema for Vertices and Edges

Before you can input your schema, you must create a space and use it. In this example we create a **nba** space and use it. We create two tags and two edge types with the following commands:

```ngql
nebula> CREATE TAG player (name string, age int);

nebula> CREATE TAG team (name string);

nebula> CREATE EDGE serve (start_year int, end_year int);

nebula> CREATE EDGE follow (degree, int);
```

## Preparing Your Configuration File

You must configure the `.yaml` configuration file, which regulates how data is organized in the CSV files. In this example, we create a `config.yaml` file.

In this example, we configure the `config.yaml` configuration file as follows:

```ngql
version: v1rc1
description: example
clientSettings:
  concurrency: 2 # number of graph clients
  channelBufferSize: 1
  space: nba
  connection:
    user: user
    password: password
    address: 127.0.0.1:3699
logPath: ./err/test.log
files:
  - path: /home/nebula/serve.csv
    failDataPath: ./err/serve.csv
    batchSize: 2
    type: csv
    csv:
      withHeader: false
      withLabel: false
    schema:
      type: edge
      edge:
        name: serve
        withRanking: false
        props:
          - name: start_year
            type: int
          - name: end_year
            type: int
  - path: /home/nebula/follow.csv
    failDataPath: ./err/follow.csv
    batchSize: 2
    type: csv
    csv:
      withHeader: false
      withLabel: false
    schema:
      type: edge
      edge:
        name: follow
        withRanking: false
        props:
          - name: degree
            type: int
  - path: /home/nebula/player.csv
    failDataPath: ./err/player.csv
    batchSize: 2
    type: csv
    csv:
      withHeader: false
      withLabel: false
    schema:
      type: vertex
      vertex:
        tags:
          - name: player
            props:
              - name: name
                type: string
              - name: age
                type: int
  - path: /home/nebula/team.csv
    failDataPath: ./err/team.csv
    batchSize: 2
    type: csv
    csv:
      withHeader: false
      withLabel: false
    schema:
      type: vertex
      vertex:
        tags:
          - name: team
            props:
              - name: name
                type: string

```

**Note**:

* In the above configuration file, you must change the IP address and the port number to yours.

* You must change the directory of the CSV files to yours, otherwise, **Nebula Importer** cannot find the CSV files.

## Preparing the CSV Data

In this example, we prepare four CSV data files: `player.csv`, `team.csv`, `serve.csv`, and `follow.csv`.

The data in the `serve.csv` file is as follows:

```csv
100,200,1997,2016
101,201,1999,2018
102,203,2006,2015
102,204,2015,2019
103,204,2017,2019
104,200,2007,2009
```

The data in the `follow.csv` file is as follows:

```csv
100,101,95
100,102,90
101,100,95
102,101,75
102,100,75
103,102,70
104,101,50
104,105,60
105,104,83
```

The data in the `player.csv` file is as follows:

```csv
100,Tim Duncan,42
101,Tony Parker,36
102,LaMarcus Aldridge,33
103,Rudy Gay,32
104,Marco Belinelli,32
105,Danny Green,31
106,Kyle Anderson,25
107,Aron Baynes,32
108,Boris Diaw,36
```

The data in the `team.csv` file is as follows:

```csv
200,Warriors
201,Nuggets
202,Rockets
203,Trail
204,Spurs
205,Thunders
206,Jazz
207,Clippers
208,Kings
```

**Note**:

* In the **serve** and **follow** CSV files, the first column is the source vertex ID, the second column is the destination vertex ID, and the other columns are consistent with the `config.yaml` file.

* In the **player** and **team** CSV files, the first column is the vertex ID and the other columns are consistent with the `config.yaml` file.

## Importing the CSV Data

After all the previous four steps are complete, you can import the CSV data with `Docker` or `Go`.

### Importing the CSV Data With Go

Before you import CSV data with `Go`, you must ensure `Go` is installed and the environment variable for `Go` is configured.

You can import the CSV data by the following steps:

1. Change your current directory to the directory where the `import.go` file is located by the following command:

```bash
$ cd /home/nebula/nebula-importer/cmd
```

2. Execute the following command to import the CSV data:

```bash
$ go run importer.go --config /home/nebula/config.yaml
```

**Note**: You must change the directory for the `import.go` file and the directory for the `config.yaml` file to yours, otherwise, the importing operation might fail.

### Importing the CSV Data With Docker

Before you import the CSV data with `Docker`, you must ensure that `Docker` is up and running.

You can import the CSV data with `Docker` by the following command:

```bash
$ sudo docker run --rm -ti --network=host \
-v /home/nebula/config.yaml:/home/nebula/config.yaml \
-v /home/nebula/:/home/nebula/ vesoft/nebula-importer \
--config /home/nebula/config.yaml
```

**Note**: You must change the directory for the `config.yaml` file to yours, otherwise the importing operation might fail.
